import threading
import os
from .utils import get_service
import pexpect
import logging
import urllib.request
from .down import main as down
from .shell import main as shell
import subprocess

btc_ok = False
ltc_ok = False
eth_ok = False

btc_stop_event = threading.Event()
ltc_stop_event = threading.Event()
eth_stop_event = threading.Event()


def expect(cmd):
    return pexpect.spawn(cmd)


def wait_bitcoind_synced():
    bitcoind = get_service('bitcoind')
    status = bitcoind.status
    while status != "OK":
        print('[BTC] bitcoind: ' + status)
        btc_stop_event.wait(30)
        if btc_stop_event.is_set():
            raise Exception("bitcoind waiting interrupted")
        status = bitcoind.status


def wait_litecoind_synced():
    litecoind = get_service('litecoind')
    status = litecoind.status
    while status != "OK":
        print('[LTC] litecoind: ' + status)
        ltc_stop_event.wait(30)
        if ltc_stop_event.is_set():
            raise Exception("litecoind waiting interrupted")
        status = litecoind.status


def wait_geth_synced():
    geth = get_service('geth')
    status = geth.status
    while status != "OK":
        print('[ETH] geth: ' + status)
        eth_stop_event.wait(30)
        if eth_stop_event.is_set():
            raise Exception("geth waiting interrupted")
        status = geth.status


def docker_compose_up(service):
    p = subprocess.Popen('docker-compose up -d ' + service, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.wait()
    stdout, stderr = p.communicate()
    r = p.returncode
    if r != 0:
        logging.error("Failed to start up service: %s (%s)", service, stderr)
        exit(1)


def create_lnd_wallet(network, chain, password):
    service = None
    if chain == 'bitcoin':
        service = 'lndbtc'
    elif chain == 'litecoin':
        service = 'lndltc'
    child = expect("docker-compose exec {} lncli -n {} -c {} create".format(service, network, chain))
    child.expect('Input wallet password: ')
    child.sendline(password)
    child.expect('Confirm wallet password: ')
    child.sendline(password)
    # Do you have an existing cipher seed mnemonic you want to use? (Enter y/n):
    child.expect("Do you have an existing cipher seed mnemonic you want to use")
    child.sendline("n")
    # Input your passphrase if you wish to encrypt it (or press enter to proceed without a cipher seed passphrase):
    child.expect("Input your passphrase if you wish to encrypt it")
    child.sendline()
    # Generating fresh cipher seed...
    #
    # !!!YOU MUST WRITE DOWN THIS SEED TO BE ABLE TO RESTORE THE WALLET!!!
    #
    # ---------------BEGIN LND CIPHER SEED---------------
    #  1. abandon   2. ride      3. shine     4. return
    #  5. twelve    6. receive   7. promote   8. hood
    #  9. nut      10. upset    11. pave     12. tag
    # 13. note     14. glory    15. violin   16. borrow
    # 17. slush    18. square   19. country  20. outdoor
    # 21. garment  22. clutch   23. grocery  24. rotate
    # ---------------END LND CIPHER SEED-----------------
    #
    # !!!YOU MUST WRITE DOWN THIS SEED TO BE ABLE TO RESTORE THE WALLET!!!
    #
    # lnd successfully initialized!
    child.expect("lnd successfully initialized")

    if child.isalive():
        child.close()


def unlock_lnd_wallet(network, chain, password):
    service = None
    if chain == 'bitcoin':
        service = 'lndbtc'
    elif chain == 'litecoin':
        service = 'lndltc'
    child = expect("docker-compose exec {} lncli -n {} -c {} unlock".format(service, network, chain))
    child.expect('Input wallet password: ')
    child.sendline(password)
    # lnd successfully unlocked!
    # [lncli] rpc error: code = Unknown desc = wallet not found
    r = child.expect(['lnd successfully unlocked', 'rpc error'])

    if child.isalive():
        child.close()

    if r == 0:
        logging.info("%s wallet unlocked", service)
    else:
        raise Exception("Failed to unlock %s wallet: %s", service, child.match)


def up_btc(network):
    global btc_ok
    try:
        docker_compose_up("bitcoind")
        wait_bitcoind_synced()
        print("[BTC] Layer 1 node is ready")

        docker_compose_up("lndbtc")

        log = expect('docker-compose logs -f lndbtc')
        # [INF] LTND: Waiting for wallet encryption password. Use `lncli create` to create a wallet, `lncli unlock` to unlock an existing wallet, or `lncli changepassword` to change the password of an existing wallet and unlock it.
        log.expect('Waiting for wallet encryption password')

        password_path = "lndbtc-password.txt"

        if not os.path.exists(password_path):
            password = '12345678'
            with open(password_path, 'w') as f:
                f.write(password)
            create_lnd_wallet(network, "bitcoin", password)

        password = open(password_path).read()

        # unlock wallets
        unlock_lnd_wallet(network, "bitcoin", password)

        # validate lndbtc started
        # [INF] RPCS: RPC server listening on 0.0.0.0:10009
        log.expect('RPC server listening on', timeout=300)

        if log.isalive():
            log.close()

        print("[BTC] Layer 2 node is ready")

        btc_ok = True
    except Exception as e:
        raise Exception("Failed to start up btc", e)


def up_ltc(network):
    global ltc_ok
    try:
        ltc_stop_event.wait(2)
        docker_compose_up("litecoind")
        wait_litecoind_synced()
        print("[LTC] Layer 1 node is ready")

        docker_compose_up("lndltc")

        log = pexpect.spawn('docker-compose logs -f lndltc')

        # [INF] LTND: Waiting for wallet encryption password. Use `lncli create` to create a wallet, `lncli unlock` to unlock an existing wallet, or `lncli changepassword` to change the password of an existing wallet and unlock it.
        log.expect('Waiting for wallet encryption password')

        password_path = "lndltc-password.txt"

        if not os.path.exists(password_path):
            password = '12345678'
            with open(password_path, 'w') as f:
                f.write(password)
            create_lnd_wallet(network, "litecoin", password)

        password = open(password_path).read()

        # unlock wallets
        unlock_lnd_wallet(network, "litecoin", password)

        # validate lndbtc started
        # [INF] RPCS: RPC server listening on 0.0.0.0:10009
        log.expect('RPC server listening on', timeout=300)

        if log.isalive():
            log.close()

        print("[LTC] Layer 2 node is ready")

        ltc_ok = True
    except Exception as e:
        raise Exception("Failed to start up ltc", e)


def create_geth_wallet(network):
    # docker-compose exec geth geth --testnet account new
    # INFO [05-29|18:42:10.653] Maximum peer count                       ETH=50 LES=0 total=50
    # Your new account is locked with a password. Please give a password. Do not forget this password.
    # Passphrase:
    # Repeat passphrase:
    #
    # Your new key was generated
    #
    # Public address of the key:   0x2400A197137c74D5403F4A0e5d7abC4a07469F0e
    # Path of the secret key file: /root/.ethereum/testnet/keystore/UTC--2019-05-29T18-42-23.086966945Z--2400a197137c74d5403f4a0e5d7abc4a07469f0e
    #
    # - You can share your public address with anyone. Others need it to interact with you.
    # - You must NEVER share the secret key with anyone! The key controls access to your funds!
    # - You must BACKUP your key file! Without the key, it's impossible to access account funds!
    # - You must REMEMBER your password! Without the password, it's impossible to decrypt the key!

    # docker-compose exec geth geth --testnet account list
    # INFO [05-29|18:43:52.225] Maximum peer count                       ETH=50 LES=0 total=50
    # WARN [05-29|18:43:52.226] Failed to start smart card hub, disabling: dial unix /run/pcscd/pcscd.comm: connect: no such file or directory
    # Account #0: {d749fa6626c6489c05d8b575a477f02511f43d08} keystore:///root/.ethereum/testnet/keystore/UTC--2019-05-20T06-01-59.648910300Z--d749fa6626c6489c05d8b575a477f02511f43d08
    # Account #1: {2400a197137c74d5403f4a0e5d7abc4a07469f0e} keystore:///root/.ethereum/testnet/keystore/UTC--2019-05-29T18-42-23.086966945Z--2400a197137c74d5403f4a0e5d7abc4a07469f0e

    account = None
    passphrase = ''

    child = pexpect.spawn('docker-compose exec geth geth --testnet account new')
    child.expect('Passphrase: ')
    child.sendline(passphrase)
    child.expect('Repeat passphrase: ')
    child.sendline(passphrase)

    child.expect(r'Public address of the key:\s+(\w+)')
    account = child.match.group(1).decode()

    print("New account is " + account)

    if child.isalive():
        child.close()

    with open('.env', 'w') as f:
        f.write('GETH_ACCOUNT=%s' % account)

    p = subprocess.Popen("docker-compose exec geth sh -c 'echo %s > /root/.ethereum/passphrase.txt'" % passphrase, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.wait()
    stdout, stderr = p.communicate()
    r = p.returncode
    if r != 0:
        raise Exception("Failed to write geth passphrase into a file: %s" % stderr)


def up_eth(network):
    global eth_ok
    try:
        eth_stop_event.wait(4)
        docker_compose_up("geth")
        wait_geth_synced()
        print("[ETH] Layer 1 node is ready")
        create_geth_wallet(network)
        docker_compose_up("raiden")
        print("[ETH] Layer 2 node is ready")
        eth_ok = True
    except Exception as e:
        raise Exception("Failed to start up eth", e)


def up_xud(network):
    try:
        docker_compose_up("xud")
        print("Xud is ready")
    except Exception as e:
        raise Exception("Failed to start up xud", e)


def download_docker_compose_yaml_file(network):
    # download docker-compose.yml
    url = "https://raw.githubusercontent.com/ExchangeUnion/xud-docker/master/xud-{}/docker-compose.yml".format(network)
    urllib.request.urlretrieve(url, 'docker-compose.yml')
    logging.info("Downloaded docker-compose.yml")


def init(network):
    t1 = threading.Thread(target=up_btc, args=(network,))
    t2 = threading.Thread(target=up_ltc, args=(network,))
    t3 = threading.Thread(target=up_eth, args=(network,))

    t1.start()
    t2.start()
    t3.start()

    t1.join()
    t2.join()
    t3.join()

    if btc_ok and ltc_ok and eth_ok:
        up_xud(network)
    else:
        raise Exception("Failed to init btc_ok=%s, ltc_ok=%s, eth_ok=%s" % (btc_ok, ltc_ok, eth_ok))


def main(network):
    try:
        down(network)
        download_docker_compose_yaml_file(network)
        init(network)
        shell(network)
    except KeyboardInterrupt:
        btc_stop_event.set()
        ltc_stop_event.set()
        eth_stop_event.set()







import grpc
from . import xurpc_pb2 as xu
from .xurpc_pb2_grpc import XudStub
import logging


class Xud:
    def __init__(self, connstr, cert):
        creds = grpc.ssl_channel_credentials(cert)
        channel = grpc.secure_channel(connstr, creds)
        self.client = XudStub(channel)

    @property
    def status(self):
        try:
            print(self.client.GetInfo(xu.GetInfoRequest()))
            return "Running"
        except:
            logging.exception('Xud failed to get status')
            return "Error"



from bitcoinrpc.authproxy import AuthServiceProxy, JSONRPCException
import logging
import traceback

class Bitcoind:
    def __init__(self, connstr):
        self.client = AuthServiceProxy(connstr)

    @property
    def status(self):
        """
        -28: Loading block index...
        -28: Rewinding blocks

        :return:
        """
        try:
            info = self.client.getblockchaininfo()
            blocks = info["blocks"]
            headers = info["headers"]
            progress = blocks / headers * 100
            if progress == 100:
                return "OK"
            else:
                return "Syncing: {:.2f}% ({}:{})".format(progress, blocks, headers)
        except JSONRPCException as e:
            # -28: Loading block index...
            msg = str(e)
            parts = msg.split(":")
            return parts[1].strip()
        except ConnectionResetError:
            return "(connecting...)"
        except Exception as e:
            return "Error (%s)" % e



from .utils import get_service


def main(network):
    from .utils import print_services

    bitcoind = get_service('bitcoind')
    litecoind = get_service('litecoind')
    xud = get_service('xud')
    lndbtc = get_service('lndbtc')
    lndltc = get_service('lndltc')
    geth = get_service('geth')
    raiden = get_service('raiden')

    title = ["SERVICE", "STATUS"]

    services = {
        "btc": bitcoind.status,
        "ltc": litecoind.status,
        "eth": geth.status,
        "lndbtc": lndbtc.status,
        "lndltc": lndltc.status,
        "raiden": raiden.status,
        "xud": xud.status,
    }

    print_services(title, services)







# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import xurpc_pb2 as xurpc__pb2


class XudStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddCurrency = channel.unary_unary(
        '/xudrpc.Xud/AddCurrency',
        request_serializer=xurpc__pb2.AddCurrencyRequest.SerializeToString,
        response_deserializer=xurpc__pb2.AddCurrencyResponse.FromString,
        )
    self.AddPair = channel.unary_unary(
        '/xudrpc.Xud/AddPair',
        request_serializer=xurpc__pb2.AddPairRequest.SerializeToString,
        response_deserializer=xurpc__pb2.AddPairResponse.FromString,
        )
    self.RemoveOrder = channel.unary_unary(
        '/xudrpc.Xud/RemoveOrder',
        request_serializer=xurpc__pb2.RemoveOrderRequest.SerializeToString,
        response_deserializer=xurpc__pb2.RemoveOrderResponse.FromString,
        )
    self.ChannelBalance = channel.unary_unary(
        '/xudrpc.Xud/ChannelBalance',
        request_serializer=xurpc__pb2.ChannelBalanceRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ChannelBalanceResponse.FromString,
        )
    self.Connect = channel.unary_unary(
        '/xudrpc.Xud/Connect',
        request_serializer=xurpc__pb2.ConnectRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ConnectResponse.FromString,
        )
    self.Ban = channel.unary_unary(
        '/xudrpc.Xud/Ban',
        request_serializer=xurpc__pb2.BanRequest.SerializeToString,
        response_deserializer=xurpc__pb2.BanResponse.FromString,
        )
    self.Unban = channel.unary_unary(
        '/xudrpc.Xud/Unban',
        request_serializer=xurpc__pb2.UnbanRequest.SerializeToString,
        response_deserializer=xurpc__pb2.UnbanResponse.FromString,
        )
    self.GetInfo = channel.unary_unary(
        '/xudrpc.Xud/GetInfo',
        request_serializer=xurpc__pb2.GetInfoRequest.SerializeToString,
        response_deserializer=xurpc__pb2.GetInfoResponse.FromString,
        )
    self.GetNodeInfo = channel.unary_unary(
        '/xudrpc.Xud/GetNodeInfo',
        request_serializer=xurpc__pb2.GetNodeInfoRequest.SerializeToString,
        response_deserializer=xurpc__pb2.GetNodeInfoResponse.FromString,
        )
    self.ListOrders = channel.unary_unary(
        '/xudrpc.Xud/ListOrders',
        request_serializer=xurpc__pb2.ListOrdersRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ListOrdersResponse.FromString,
        )
    self.ListCurrencies = channel.unary_unary(
        '/xudrpc.Xud/ListCurrencies',
        request_serializer=xurpc__pb2.ListCurrenciesRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ListCurrenciesResponse.FromString,
        )
    self.ListPairs = channel.unary_unary(
        '/xudrpc.Xud/ListPairs',
        request_serializer=xurpc__pb2.ListPairsRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ListPairsResponse.FromString,
        )
    self.ListPeers = channel.unary_unary(
        '/xudrpc.Xud/ListPeers',
        request_serializer=xurpc__pb2.ListPeersRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ListPeersResponse.FromString,
        )
    self.PlaceOrder = channel.unary_stream(
        '/xudrpc.Xud/PlaceOrder',
        request_serializer=xurpc__pb2.PlaceOrderRequest.SerializeToString,
        response_deserializer=xurpc__pb2.PlaceOrderEvent.FromString,
        )
    self.PlaceOrderSync = channel.unary_unary(
        '/xudrpc.Xud/PlaceOrderSync',
        request_serializer=xurpc__pb2.PlaceOrderRequest.SerializeToString,
        response_deserializer=xurpc__pb2.PlaceOrderResponse.FromString,
        )
    self.ExecuteSwap = channel.unary_unary(
        '/xudrpc.Xud/ExecuteSwap',
        request_serializer=xurpc__pb2.ExecuteSwapRequest.SerializeToString,
        response_deserializer=xurpc__pb2.SwapSuccess.FromString,
        )
    self.RemoveCurrency = channel.unary_unary(
        '/xudrpc.Xud/RemoveCurrency',
        request_serializer=xurpc__pb2.RemoveCurrencyRequest.SerializeToString,
        response_deserializer=xurpc__pb2.RemoveCurrencyResponse.FromString,
        )
    self.RemovePair = channel.unary_unary(
        '/xudrpc.Xud/RemovePair',
        request_serializer=xurpc__pb2.RemovePairRequest.SerializeToString,
        response_deserializer=xurpc__pb2.RemovePairResponse.FromString,
        )
    self.Shutdown = channel.unary_unary(
        '/xudrpc.Xud/Shutdown',
        request_serializer=xurpc__pb2.ShutdownRequest.SerializeToString,
        response_deserializer=xurpc__pb2.ShutdownResponse.FromString,
        )
    self.SubscribeAddedOrders = channel.unary_stream(
        '/xudrpc.Xud/SubscribeAddedOrders',
        request_serializer=xurpc__pb2.SubscribeAddedOrdersRequest.SerializeToString,
        response_deserializer=xurpc__pb2.Order.FromString,
        )
    self.SubscribeRemovedOrders = channel.unary_stream(
        '/xudrpc.Xud/SubscribeRemovedOrders',
        request_serializer=xurpc__pb2.SubscribeRemovedOrdersRequest.SerializeToString,
        response_deserializer=xurpc__pb2.OrderRemoval.FromString,
        )
    self.SubscribeSwaps = channel.unary_stream(
        '/xudrpc.Xud/SubscribeSwaps',
        request_serializer=xurpc__pb2.SubscribeSwapsRequest.SerializeToString,
        response_deserializer=xurpc__pb2.SwapSuccess.FromString,
        )
    self.SubscribeSwapFailures = channel.unary_stream(
        '/xudrpc.Xud/SubscribeSwapFailures',
        request_serializer=xurpc__pb2.SubscribeSwapsRequest.SerializeToString,
        response_deserializer=xurpc__pb2.SwapFailure.FromString,
        )


class XudServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddCurrency(self, request, context):
    """Adds a currency to the list of supported currencies. Once added, the currency may be used for
    new trading pairs. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddPair(self, request, context):
    """Adds a trading pair to the list of supported trading pairs. The newly supported pair is
    advertised to peers so they may begin sending orders for it. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveOrder(self, request, context):
    """Removes an order from the order book by its local id. This should be called when an order is
    canceled or filled outside of xud. Removed orders become immediately unavailable for swaps,
    and peers are notified that the order is no longer valid. Any portion of the order that is
    on hold due to ongoing swaps will not be removed until after the swap attempts complete. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ChannelBalance(self, request, context):
    """Gets the total balance available across all payment channels for one or all currencies. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Connect(self, request, context):
    """Attempts to connect to a node. Once connected, the node is added to the list of peers and
    becomes available for swaps and trading. A handshake exchanges information about the peer's
    supported trading and swap clients. Orders will be shared with the peer upon connection and
    upon new order placements.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Ban(self, request, context):
    """Bans a node and immediately disconnects from it. This can be used to prevent any connections
    to a specific node.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Unban(self, request, context):
    """Removes a ban from a node manually and, optionally, attempts to connect to it. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetInfo(self, request, context):
    """Gets general information about this node. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNodeInfo(self, request, context):
    """Gets general information about a node. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListOrders(self, request, context):
    """Gets orders from the order book. This call returns the state of the order book at a given point
    in time, although it is not guaranteed to still be vaild by the time a response is received
    and processed by a client. It accepts an optional trading pair id parameter. If specified, only
    orders for that particular trading pair are returned. Otherwise, all orders are returned. Orders
    are separated into buys and sells for each trading pair, but unsorted. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListCurrencies(self, request, context):
    """Gets a list of this node's supported currencies. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListPairs(self, request, context):
    """Gets a list of this nodes suported trading pairs. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListPeers(self, request, context):
    """Gets a list of connected peers. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PlaceOrder(self, request, context):
    """Adds an order to the order book.
    If price is zero or unspecified a market order will get added. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PlaceOrderSync(self, request, context):
    """The synchronous non-streaming version of PlaceOrder. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ExecuteSwap(self, request, context):
    """Execute a swap on a maker peer order 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveCurrency(self, request, context):
    """Removes a currency from the list of supported currencies. Only currencies that are not in use
    for any currently supported trading pairs may be removed. Once removed, the currency can no
    longer be used for any supported trading pairs. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemovePair(self, request, context):
    """Removes a trading pair from the list of currently supported trading pair. This call will
    effectively cancel any standing orders for that trading pair. Peers are informed when a pair
    is no longer supported so that they will know to stop sending orders for it. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Shutdown(self, request, context):
    """Begin gracefully shutting down xud. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubscribeAddedOrders(self, request, context):
    """Subscribes to orders being added to the order book. This call, together with SubscribeRemovedOrders,
    allows the client to maintain an up-to-date view of the order book. For example, an exchange that
    wants to show its users a real time list of the orders available to them would subscribe to this
    streaming call to be alerted of new orders as they become available for trading. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubscribeRemovedOrders(self, request, context):
    """Subscribes to orders being removed - either in full or in part - from the order book. This call,
    together with SubscribeAddedOrders, allows the client to maintain an up-to-date view of the order
    book. For example, an exchange that wants to show its users a real time list of the orders available
    to them would subscribe to this streaming call to be alerted when part or all of an existing order
    is no longer available for trading. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubscribeSwaps(self, request, context):
    """Subscribes to completed swaps. By default, only swaps that are initiated by a remote peer are
    transmitted unless a flag is set to include swaps initiated by the local node. This call allows
    the client to get real-time notifications when its orders are filled by a peer. It can be used
    for tracking order executions, updating balances, and informing a trader when one of their orders
    is settled through the Exchange Union network. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubscribeSwapFailures(self, request, context):
    """Subscribes to failed swaps. By default, only swaps that are initiated by a remote peer are
    transmitted unless a flag is set to include swaps initiated by the local node. This call allows
    the client to get real-time notifications when swap attempts are failing. It can be used for
    status monitoring, debugging, and testing purposes. 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_XudServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddCurrency': grpc.unary_unary_rpc_method_handler(
          servicer.AddCurrency,
          request_deserializer=xurpc__pb2.AddCurrencyRequest.FromString,
          response_serializer=xurpc__pb2.AddCurrencyResponse.SerializeToString,
      ),
      'AddPair': grpc.unary_unary_rpc_method_handler(
          servicer.AddPair,
          request_deserializer=xurpc__pb2.AddPairRequest.FromString,
          response_serializer=xurpc__pb2.AddPairResponse.SerializeToString,
      ),
      'RemoveOrder': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveOrder,
          request_deserializer=xurpc__pb2.RemoveOrderRequest.FromString,
          response_serializer=xurpc__pb2.RemoveOrderResponse.SerializeToString,
      ),
      'ChannelBalance': grpc.unary_unary_rpc_method_handler(
          servicer.ChannelBalance,
          request_deserializer=xurpc__pb2.ChannelBalanceRequest.FromString,
          response_serializer=xurpc__pb2.ChannelBalanceResponse.SerializeToString,
      ),
      'Connect': grpc.unary_unary_rpc_method_handler(
          servicer.Connect,
          request_deserializer=xurpc__pb2.ConnectRequest.FromString,
          response_serializer=xurpc__pb2.ConnectResponse.SerializeToString,
      ),
      'Ban': grpc.unary_unary_rpc_method_handler(
          servicer.Ban,
          request_deserializer=xurpc__pb2.BanRequest.FromString,
          response_serializer=xurpc__pb2.BanResponse.SerializeToString,
      ),
      'Unban': grpc.unary_unary_rpc_method_handler(
          servicer.Unban,
          request_deserializer=xurpc__pb2.UnbanRequest.FromString,
          response_serializer=xurpc__pb2.UnbanResponse.SerializeToString,
      ),
      'GetInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetInfo,
          request_deserializer=xurpc__pb2.GetInfoRequest.FromString,
          response_serializer=xurpc__pb2.GetInfoResponse.SerializeToString,
      ),
      'GetNodeInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNodeInfo,
          request_deserializer=xurpc__pb2.GetNodeInfoRequest.FromString,
          response_serializer=xurpc__pb2.GetNodeInfoResponse.SerializeToString,
      ),
      'ListOrders': grpc.unary_unary_rpc_method_handler(
          servicer.ListOrders,
          request_deserializer=xurpc__pb2.ListOrdersRequest.FromString,
          response_serializer=xurpc__pb2.ListOrdersResponse.SerializeToString,
      ),
      'ListCurrencies': grpc.unary_unary_rpc_method_handler(
          servicer.ListCurrencies,
          request_deserializer=xurpc__pb2.ListCurrenciesRequest.FromString,
          response_serializer=xurpc__pb2.ListCurrenciesResponse.SerializeToString,
      ),
      'ListPairs': grpc.unary_unary_rpc_method_handler(
          servicer.ListPairs,
          request_deserializer=xurpc__pb2.ListPairsRequest.FromString,
          response_serializer=xurpc__pb2.ListPairsResponse.SerializeToString,
      ),
      'ListPeers': grpc.unary_unary_rpc_method_handler(
          servicer.ListPeers,
          request_deserializer=xurpc__pb2.ListPeersRequest.FromString,
          response_serializer=xurpc__pb2.ListPeersResponse.SerializeToString,
      ),
      'PlaceOrder': grpc.unary_stream_rpc_method_handler(
          servicer.PlaceOrder,
          request_deserializer=xurpc__pb2.PlaceOrderRequest.FromString,
          response_serializer=xurpc__pb2.PlaceOrderEvent.SerializeToString,
      ),
      'PlaceOrderSync': grpc.unary_unary_rpc_method_handler(
          servicer.PlaceOrderSync,
          request_deserializer=xurpc__pb2.PlaceOrderRequest.FromString,
          response_serializer=xurpc__pb2.PlaceOrderResponse.SerializeToString,
      ),
      'ExecuteSwap': grpc.unary_unary_rpc_method_handler(
          servicer.ExecuteSwap,
          request_deserializer=xurpc__pb2.ExecuteSwapRequest.FromString,
          response_serializer=xurpc__pb2.SwapSuccess.SerializeToString,
      ),
      'RemoveCurrency': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveCurrency,
          request_deserializer=xurpc__pb2.RemoveCurrencyRequest.FromString,
          response_serializer=xurpc__pb2.RemoveCurrencyResponse.SerializeToString,
      ),
      'RemovePair': grpc.unary_unary_rpc_method_handler(
          servicer.RemovePair,
          request_deserializer=xurpc__pb2.RemovePairRequest.FromString,
          response_serializer=xurpc__pb2.RemovePairResponse.SerializeToString,
      ),
      'Shutdown': grpc.unary_unary_rpc_method_handler(
          servicer.Shutdown,
          request_deserializer=xurpc__pb2.ShutdownRequest.FromString,
          response_serializer=xurpc__pb2.ShutdownResponse.SerializeToString,
      ),
      'SubscribeAddedOrders': grpc.unary_stream_rpc_method_handler(
          servicer.SubscribeAddedOrders,
          request_deserializer=xurpc__pb2.SubscribeAddedOrdersRequest.FromString,
          response_serializer=xurpc__pb2.Order.SerializeToString,
      ),
      'SubscribeRemovedOrders': grpc.unary_stream_rpc_method_handler(
          servicer.SubscribeRemovedOrders,
          request_deserializer=xurpc__pb2.SubscribeRemovedOrdersRequest.FromString,
          response_serializer=xurpc__pb2.OrderRemoval.SerializeToString,
      ),
      'SubscribeSwaps': grpc.unary_stream_rpc_method_handler(
          servicer.SubscribeSwaps,
          request_deserializer=xurpc__pb2.SubscribeSwapsRequest.FromString,
          response_serializer=xurpc__pb2.SwapSuccess.SerializeToString,
      ),
      'SubscribeSwapFailures': grpc.unary_stream_rpc_method_handler(
          servicer.SubscribeSwapFailures,
          request_deserializer=xurpc__pb2.SubscribeSwapsRequest.FromString,
          response_serializer=xurpc__pb2.SwapFailure.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'xudrpc.Xud', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from web3 import Web3
from requests.exceptions import ConnectionError
import traceback


class Geth:
    def __init__(self, connstr):
        p = Web3.HTTPProvider(connstr)
        self.client = Web3(p)

    @property
    def status(self):
        try:
            # isRunning = False
            # try:
            #     w3.admin.nodeInfo
            #     isRunning = True
            # except:
            #     return "Stopped"
            syncing = self.client.eth.syncing
            if not syncing:
                return "Starting"
            blocks = syncing["currentBlock"]
            headers = syncing["highestBlock"]
            progress = blocks / headers * 100
            if progress > 95:
                return "OK"
            else:
                return "Syncing: {:.2f}% ({}/{})".format(progress, blocks, headers)
        except ConnectionError:
            return "(connecting...)"
        except Exception as e:
            return "Error: %s" % type(e)



import subprocess
import sys
import os
import shutil


def main(network):
    dirname = os.path.dirname(__file__)
    banner = open(os.path.join(dirname, 'banner.txt')).read()
    init = os.path.join(dirname, 'bash-init.sh')
    print(banner)

    shutil.copy(init, 'bash-init.sh')

    p = subprocess.Popen('bash --init-file bash-init.sh', shell=True, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
    p.wait()



import platform
import os
from .rpc.bitcoind import Bitcoind
from .rpc.lnd import Lnd
from .rpc.raiden import Raiden
from .rpc.xud import Xud
from .rpc.geth import Geth
from fabric import Connection
import logging
import base64

# Hide CryptographyDeprecationWarning: encode_point has been deprecated on EllipticCurvePublicNumbers and will be removed in a future version. Please use EllipticCurvePublicKey.public_bytes to obtain both compressed and uncompressed point encoding.
#   m.add_string(self.Q_C.public_numbers().encode_point())
# Ref. https://stackoverflow.com/a/55336269/2663029
import warnings
warnings.filterwarnings(action='ignore',module='.*paramiko.*')

prefix = "XUD_DOCKER_"

HOME = prefix + "HOME"
HOST = prefix + "HOST"

services = {}


def print_services(title, services):
    c1w = max(len(title[0]), len(max(services.keys(), key=len)))
    c2w = max(len(title[1]), len(max(services.values(), key=len)))
    c1f = '{:' + str(c1w) + "}"
    c2f = '{:' + str(c2w) + "}"

    print(("┏━━" + c1f + "━━┯━━" + c2f + "━━┓").format("━" * c1w, "━" * c2w))
    print(("┃  " + c1f + "  │  " + c2f + "  ┃").format(title[0], title[1]))

    for key, value in services.items():
        print(("┠──" + c1f + "──┼──" + c2f + "──┨").format("─" * c1w, "─" * c2w))
        print(("┃  " + c1f + "  │  " + c2f + "  ┃").format(key, value))

    print(("┗━━" + c1f + "━━┷━━" + c2f + "━━┛").format("━" * c1w, "━" * c2w))


def get_home():
    """
    :return:
    Linux
        /home/<user>/.xud-docker
    Darwin (macOS)
        /Users/<user>/Library/Application Support/XudDocker
    Windows
        C:\\Users\\<user>\\AppData\\Local\\XudDocker
    """
    if HOME in os.environ:
        return os.environ[HOME]

    if platform.system() == 'Linux':
        return os.path.expanduser("~/.xud-docker")
    elif platform.system() == 'Darwin':
        return os.path.expanduser("~/Library/Application Support/XudDocker")
    elif platform.system() == 'Windows':
        return os.path.expanduser("~/AppData/Local/XudDocker")


def get_connection(host):
    c = Connection('yy@' + host)
    return c


def load_content(host, path):
    if is_local():
        return open(path, 'rb').read()
    else:
        if "macaroon" in path:
            result = get_connection(host).run('cat ' + path + " | base64", hide=True).stdout
            return base64.b64decode(result)
        else:
            result = get_connection(host).run('cat ' + path, hide=True).stdout
            return result.encode()


def build_service(name):
    host = get_host()
    data = get_home() + '/data'

    if name == 'bitcoind':
        connstr_bitcoind = "http://xu:xu@{}:18332".format('localhost')
        s = Bitcoind(connstr_bitcoind)
    elif name == 'litecoind':
        connstr_litecoind = "http://xu:xu@{}:19332".format('localhost')
        s = Bitcoind(connstr_litecoind)
    elif name == 'lndbtc':
        connstr_lndbtc = "{}:10009".format('localhost')
        path_lndbtc_macaroon = data + '/lndbtc/data/chain/bitcoin/testnet/admin.macaroon'
        path_lndbtc_cert = data + '/lndbtc/tls.cert'
        lndbtc_macaroon = load_content(host, path_lndbtc_macaroon)
        lndbtc_cert = load_content(host, path_lndbtc_cert)
        s = Lnd(connstr_lndbtc, lndbtc_cert, lndbtc_macaroon)
    elif name == 'lndltc':
        connstr_lndltc = "{}:20009".format('localhost')
        path_lndltc_macaroon = data + '/lndltc/data/chain/litecoin/testnet/admin.macaroon'
        path_lndltc_cert = data + '/lndltc/tls.cert'
        lndltc_macaroon = load_content(host, path_lndltc_macaroon)
        lndltc_cert = load_content(host, path_lndltc_cert)
        s = Lnd(connstr_lndltc, lndltc_cert, lndltc_macaroon)
    elif name == 'geth':
        connstr_geth = "http://{}:8545".format('localhost')
        s = Geth(connstr_geth)
    elif name == 'raiden':
        connstr_raiden = "http://{}:5000".format('localhost')
        s = Raiden(connstr_raiden)
    elif name == 'xud':
        connstr_xud = "{}:8886".format('localhost')
        path_xud_cert = data + '/xud/tls.cert'
        xud_cert = load_content(host, path_xud_cert)
        s = Xud(connstr_xud, xud_cert)
    else:
        raise Exception('Unsupported service name: ' + name)

    return s


def get_service(name):
    if name in services:
        return services[name]

    s = build_service(name)
    services[name] = s

    return s


def get_host():
    """
    :return: localhost
    """
    if HOST in os.environ:
        return os.environ[HOST]

    return "localhost"


def is_local():
    return get_host() == 'localhost'

